// Boost.Geometry

// Copyright (c) 2021, Oracle and/or its affiliates.

// Contributed and/or modified by Vissarion Fysikopoulos, on behalf of Oracle

// Licensed under the Boost Software License version 1.0.
// http://www.boost.org/users/license.html

#ifndef BOOST_GEOMETRY_ALGORITHMS_CLOSEST_POINTS_HPP
#define BOOST_GEOMETRY_ALGORITHMS_CLOSEST_POINTS_HPP

#include <boost/geometry/algorithms/detail/closest_points/interface.hpp>
#include <boost/geometry/algorithms/detail/closest_points/implementation.hpp>

#endif // BOOST_GEOMETRY_ALGORITHMS_CLOSEST_POINTS_HPP
